/*
 * PE Resource Zero-out Utility
 *
 * (C) Copyright 2012 - Shao Miller <sha0.miller@gmail.com>
 *                      All Rights Reserved
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, Inc., 53 Temple Place Ste 330,
 *   Boston MA 02111-1307, USA; either version 2 of the License, or
 *   (at your option) any later version; incorporated herein by reference.
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

static const char * check_opt(char ** argv, const char * optname);
static int usage(char * progname);
static int zerores(const char * filename, const char * reslist);
static int examine_pe(const char * filename);

int main(int argc, char ** argv) {
    const char blurb[] =
"\n"
"  PE Resource Zero-out Utility\n"
"                    December 4th, 2012 - Shao Miller <sha0.miller@gmail.com>\n"
"                    GNU General Public License 2 (or later)\n\n";
    char ** args;
    int show_usage;
    const char * filename;
    const char * reslist;
    const char * tmp;

    printf(blurb);

    /* Process command-line options */
    show_usage = argc;
    filename = NULL;
    reslist = NULL;
    for (args = argv + 1; *args; ++args) {
        tmp = check_opt(args, "--pe");
        if (tmp) {
            if (filename || !*tmp) {
                show_usage = 1;
                break;
              }
            show_usage = 0;
            filename = tmp;
            ++args;
            continue;
          }

        tmp = check_opt(args, "--reslist");
        if (tmp) {
            if (reslist || !*tmp) {
                show_usage = 1;
                break;
              }
            show_usage = 0;
            reslist = tmp;
            ++args;
            continue;
          }

        /* Invalid option */
        show_usage = 1;
        break;
      }

    if (reslist && !filename) {
        fprintf(stderr, "\n---NOTICE---: --reslist requires --pe\n");
        show_usage = 1;
      }

    /* Show the usage? */
    if (show_usage)
      return usage(argv[0]);

    if (filename)
      printf("Using PE filename: %s\n", filename);

    if (reslist)
      printf("Resource list provided: %s\n", reslist);

    if (reslist)
      return zerores(filename, reslist);

    if (filename)
      return examine_pe(filename);

    printf("Done\n");
    return EXIT_SUCCESS;
  }

static const char * check_opt(char ** argv, const char * optname) {
    size_t optlen;

    if (!argv || !optname || !argv[0])
      return NULL;

    optlen = strlen(optname) + 1;

    if (!strncmp(argv[0], optname, optlen))
      return argv[1] ? argv[1] : "";

    return NULL;
  }

static int usage(char * progname) {
    const char usage_text[] =
"  Usage: %s [<opt> [...]]\n";
    const char * options_text[] = {
"\n"
"  Where each <opt> may be one of those below, specified once, each:\n"
"\n",
"  --pe <filename>\n"
"    Specifies that <filename> is the PE file to work with\n"
"\n",
"  --reslist [<id>[,<additional_id>[...]]]\n"
"    Specifies that the list of one or more integer IDs have associated\n"
"    resources, and those resources should be zeroed.  <id> has the form\n"
"    #<type>:#<name> and both of those parameters are integers\n"
"\n",
    NULL
  };
    const char ** option_text;

    fprintf(stderr, usage_text, progname);
    for (option_text = options_text; *option_text; ++option_text)
      fprintf(stderr, *option_text);

    return EXIT_FAILURE;
  }

static int internal_error(void) {
    fprintf(stderr, "\n---NOTICE---: Internal error!\n");
    return EXIT_FAILURE;
  }

#include <windows.h>

struct s_res_context {
    HANDLE module_upd;
    size_t res_sz;
    VOID * replacement;
  };

static int show_err(const char * msg);
static int zero_one_res(const char * filename, char * res_id);
static BOOL CALLBACK enum_res_types(
    HMODULE module,
    CHAR * res_type,
    LONG_PTR context
  );
static BOOL CALLBACK enum_res_name(
    HMODULE module,
    const CHAR * res_type,
    CHAR * res_name,
    LONG_PTR context
  );
static BOOL CALLBACK zero_one_language(
    HMODULE module,
    const CHAR * res_type,
    const CHAR * res_name,
    WORD lang,
    LONG_PTR context
  );

static int zerores(const char * filename, const char * reslist) {
    HMODULE module;
    int rv;
    char res_id[80];
    unsigned int res_count;
    size_t len;
    const char * next_res_id;

    if (!filename || !reslist)
      return internal_error();

    module = LoadLibraryA(filename);
    if (!module) {
        rv = show_err("Couldn't load PE!\n");
        goto err_module;
      }
    FreeLibrary(module);

    rv = EXIT_SUCCESS;

    for (res_count = 0; reslist; ++res_count) {
        len = sizeof res_id - 1;
        next_res_id = strchr(reslist, ',');
        if (next_res_id && (size_t) (next_res_id - reslist) < len)
          len = next_res_id - reslist;
          
        strncpy(res_id, reslist, len);
        res_id[len] = '\0';
        rv = zero_one_res(filename, res_id);
        if (rv != EXIT_SUCCESS) {
            rv = show_err("Error zeroing resource.  Stopping...\n");
            break;
          }

        reslist = next_res_id ? next_res_id + 1 : NULL;
      }

    printf("Processed %u resources\n", res_count);

    FreeLibrary(module);
    err_module:

    return rv;
  }

static int examine_pe(const char * filename) {
    HMODULE module;
    int rv;
    BOOL ok;

    if (!filename)
      return internal_error();

    module = LoadLibraryA(filename);
    if (!module) {
        rv = show_err("Couldn't load PE!\n");
        goto err_module;
      }

    ok = EnumResourceTypesA(module, enum_res_types, 0);
    if (!ok) {
        rv = show_err("Error while enumerating resources!\n");
        goto err_res_types;
      }

    printf("Done\n");
    rv = EXIT_SUCCESS;

    err_res_types:

    FreeLibrary(module);
    err_module:

    return rv;
  }

static int show_err(const char * msg) {
    DWORD err;
    CHAR * buf;

    err = GetLastError();
    if (!err)
      return EXIT_FAILURE;

    FormatMessageA(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        err,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (CHAR *) &buf,
        0,
        NULL
      );

    fprintf(stderr, "\n---NOTICE---: %s\n%s", msg, buf);

    return EXIT_FAILURE;
  }

static int zero_one_res(const char * filename, char * res_id) {
    HMODULE module;
    int rv;
    BOOL discard;
    struct s_res_context res_context;
    char * sep;
    char * res_type;
    char * res_name;
    HRSRC res;

    /* Assume failure */
    rv = EXIT_FAILURE;
    discard = TRUE;

    sep = strchr(res_id, ':');
    if (!sep) {
        fprintf(stderr, "Invalid resource ID provided.  See usage\n");
        SetLastError(0);
        goto err_sep;
      }

    *sep = '\0';
    res_type = res_id;
    res_name = sep + 1;

    res_context.module_upd = BeginUpdateResourceA(filename, FALSE);
    if (!res_context.module_upd) {
        fprintf(stderr, "Unable to load PE for update!\n");
        goto err_module_upd;
      }

    module = LoadLibraryA(filename);
    if (!module) {
        rv = show_err("Couldn't load PE!\n");
        goto err_module;
      }

    printf("Trying to zero resource: Type: %s Name: %s\n", res_type, res_name);
    res = FindResourceA(module, res_name, res_type);
    if (!res) {
        fprintf(stderr, "Unable to load resource\n");
        goto err_res;
      }

    res_context.res_sz = SizeofResource(module, res);
    if (!res_context.res_sz) {
        fprintf(stderr, "Resource has size zero\n");
        goto err_res_sz;
      }

    res_context.replacement = calloc(1, res_context.res_sz);
    if (!res_context.replacement) {
        fprintf(stderr, "Unable to allocate replacement data\n");
        goto err_replacement;
      }

    discard = !EnumResourceLanguagesA(
        module,
        res_type,
        res_name,
        zero_one_language,
        (LONG_PTR) &res_context
      );

    rv = EXIT_SUCCESS;

    free(res_context.replacement);
    err_replacement:

    err_res_sz:

    err_res:

    FreeLibrary(module);
    err_module:

    discard = !EndUpdateResourceA(res_context.module_upd, discard);
    if (discard)
      rv = EXIT_FAILURE;
    err_module_upd:

    err_sep:

    return rv;
  }

static BOOL CALLBACK enum_res_types(
    HMODULE module,
    CHAR * res_type,
    LONG_PTR context
  ) {
    return EnumResourceNamesA(module, res_type, enum_res_name, context);
  }

static BOOL CALLBACK enum_res_name(
    HMODULE module,
    const CHAR * res_type,
    CHAR * res_name,
    LONG_PTR context
  ) {
    WORD res_type_num;
    WORD res_name_num;

    (VOID) module;
    (VOID) context;

    printf("Found resource: ");

    printf("Type: ");
    if (IS_INTRESOURCE(res_type)) {
        res_type_num = (DWORD) res_type;
        printf("#%d ", res_type_num);
      } else {
        printf("%s ", res_type);
      }

    printf("Name: ");
    if (IS_INTRESOURCE(res_name)) {
        res_name_num = (DWORD) res_name;
        printf("#%d ", res_name_num);
      } else {
        printf("%s ", res_name);
      }

    printf("\n");
    return TRUE;
  }

static BOOL CALLBACK zero_one_language(
    HMODULE module,
    const CHAR * res_type,
    const CHAR * res_name,
    WORD lang,
    LONG_PTR context
  ) {
    struct s_res_context * res_context;

    (VOID) module;

    res_context = (VOID *) context;

    printf("Trying to zero resource language: %u\n", lang);
    return UpdateResourceA(
        res_context->module_upd,
        *res_type == '#' ? MAKEINTRESOURCE(atoi(res_type + 1)) : res_type,
        *res_name == '#' ? MAKEINTRESOURCE(atoi(res_name + 1)) : res_name,
        lang,
        res_context->replacement,
        res_context->res_sz
      );
  }

